unit tn38121;
{ Telnet 3812 emulation (C) 1999, by Paul TOTH <tothpaul@multimania.com>
  http://www.multimania.com/tothpaul

  New Version, because some printer drivers doesn't support the Escape() API
  I'm using now WritePrinter() API :)
}
{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

interface

{-$DEFINE STREAM - save the data flow to PRINT*.PCL }
{$DEFINE PRINT - print out the file to specified printer }

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  sockets, StdCtrls, ComCtrls, ExtCtrls, Menus;

type
  THead5250=packed record
   Size:word; // head + data
   Kind:word; // $A012 - WARNING ! can change in future version
   Flow:word; // Data Flow
  end;

  TForm1 = class(TForm)
    Telnet: TSocket;
    Start: TButton;
    About: TButton;
    OnLine: TShape;
    DataFlow: TShape;
    More: TButton;
    Label8: TLabel;
    Printers: TComboBox;
    Panel: TGroupBox;
    Label1: TLabel;
    HostName: TEdit;
    Label2: TLabel;
    MSGQName: TEdit;
    MSGQLib: TEdit;
    PPS1: TComboBox;
    Label3: TLabel;
    PPS2: TComboBox;
    Label4: TLabel;
    Label5: TLabel;
    ENV: TComboBox;
    Label7: TLabel;
    TypeModel: TComboBox;
    Label6: TLabel;
    WSCSTName: TEdit;
    WSCSTLib: TEdit;
    Device: TLabel;
    DevName: TEdit;
    PrinterStatus: TLabel;
    procedure TelnetConnect(Sender: TObject; ASocket: Integer);
    procedure TelnetRead(Sender: TObject; ASocket: Integer);
    procedure TelnetClose(Sender: TObject; ASocket: Integer);
    procedure TelnetError(Sender: TObject; Error: Word; ASocket: Integer;
      Event: Word);
    procedure FormCreate(Sender: TObject);
    procedure StartClick(Sender: TObject);
    procedure AboutClick(Sender: TObject);
    procedure MoreClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    { Dclarations prives }
    Status:(stDATA,stIAC,stDO,stSB,stWILL,stHEAD);
    Data:string;
    SubNegoce:String;

    Head:THead5250;
    HeadSize:word;
    DataSize:word;

    Data2:string;

    number:integer;
    ibmcode:byte;
    {$IFDEF STREAM}
     Stream:TMemoryStream;
    {$ENDIF}
    {$IFDEF PRINT}
    hprinter:THandle;
    hpcl:string;
    {$ENDIF}

    procedure StopClick(Sender: TObject);
    procedure LessClick(Sender: TObject);
    procedure ListPrinters;
    procedure SetStatus(s:string);
    property PrinterStatusCaption:string write SetStatus;
    procedure wmToolTrayIcon(Var msg:TMessage); message WM_USER;
  public
    { Dclarations publiques }
    procedure Error(Msg:string);
    procedure ExFmt;
  end;

var
  Form1: TForm1;

implementation

uses EBCDIC,WinSpool;

{$R *.DFM}
{$IFDEF PRINT}
function OpenPrinter(Printer:string):THandle;
 begin
  if not WinSpool.OpenPrinter(PChar(Printer),Result,nil) then Result:=0;
 end;

function BeginDoc(Handle:THandle;Title:string):boolean;
 var
  DocInfo:record
   DocName :PChar;
   OutFile :PChar;
   DataType:PChar;
  end;
 begin
  with DocInfo do begin
   DocName :=PChar(Title);
   OutFile :=nil;
   DataType:=nil;
  end;
  Result:=(StartDocPrinter(Handle,1,@DocInfo)<>0);
 end;

function WritePrinter(Handle:THandle;s:string):boolean;
 var
  Size:DWORD;
 begin
  Result:=WinSpool.WritePrinter(Handle,@s[1],Length(s),Size) and (Size=Length(s));
 end;

procedure EndDoc(var Handle:THandle);
 begin
  if Handle<>0 then begin
   ClosePrinter(Handle);
   Handle:=0;
  end;
 end;

procedure AbortDoc(var Handle:THandle);
 begin
  if Handle<>0 then begin
   AbortPrinter(Handle);
   EndDoc(Handle);
  end;
 end;
{$ENDIF}

procedure TForm1.TelnetConnect(Sender: TObject; ASocket: Integer);
begin
 OnLine.Brush.Color:=clLime;
 Start.Caption:='Stop';
 Start.OnClick:=StopClick;
 Start.Enabled:=True;
 Status:=stDATA;
 Data:='';
end;

procedure TForm1.TelnetRead(Sender: TObject; ASocket: Integer);
Const
{ RFC 854
   In summary, WILL XXX is sent, by either party, to indicate that
   party's desire (offer) to begin performing option XXX, DO XXX and
   DON'T XXX being its positive and negative acknowledgments; similarly,
   DO XXX is sent to indicate a desire (request) that the other party
   (i.e., the recipient of the DO) begin performing option XXX, WILL XXX
   and WON'T XXX being the positive and negative acknowledgments.  Since
   the NVT is what is left when no options are enabled, the DON'T and
   WON'T responses are guaranteed to leave the connection in a state
   which both ends can handle.  Thus, all hosts may implement their
   TELNET processes to be totally unaware of options that are not
   supported, simply returning a rejection to (i.e., refusing) any
   option request that cannot be understood.
}
 DEBUG = TRUE;
 TERM  = 'IBM-3812-1';

 tnEND      =#239; // End of 5250 DataStream
 tnSE       =#240;
 tnNOP      =#241;
 tnDATA     =#242;
 tnBRK      =#243;
 tnINT      =#244;
 tnABORT    =#245;
 tnAYT      =#246; // Are you there
 tnERASECHAR=#247;
 tnERASELINE=#248;
 tnGOAHEAD  =#249;
 tnSB       =#250;
 tnWILL     =#251;
 tnWONT     =#252;
 tnDO       =#253;
 tnDONT     =#254;
 tnIAC      =#255;

 tnBIN        =#0;
 tnECHO       =#1;
 tnSUPPRESS_GA=#3;
 tnTIMING     =#6;

 tnTERMTYPE   =#24; // Terminal type
 tnEOR        =#25;
 tnNEWENV     =#39; // Session name

 ttIS      =#0;
 ttSEND    =#1;
 ttINFO    =#2;

 tnVAR     =#0;
 tnVAL     =#1;

 function PaperSource(List:TComboBox):string;
  begin
   case List.ItemIndex of
    0:Result:=#$FF#$FF;
    1:Result:=#$00;
    2:Result:=#$02#$01;
    3:Result:=#$02#$02;
    4:Result:=#$02#$03;
    5:Result:=#$04;
    6:Result:=#$05;
    7:Result:=#$06;
    8:Result:=#$07;
    9:Result:=#$08;
   10:Result:=#$0E;
   11:Result:=#$0F;
   12:Result:=#$10;
   else Result:=#$FF#$FF;
   end;
  end;

 function EnvSource(List:TComboBox):string;
  begin
   case List.ItemIndex of
    0:Result:=#$FF#$FF;
    1:Result:=#$00;
    2:Result:=#$06;
    3:Result:=#$09;
    4:Result:=#$0A;
    5:Result:=#$0B;
    6:Result:=#$0C;
    7:Result:=#$0D;
   else Result:=#$FF#$FF;
   end;
  end;

const
 busy:boolean=false;
var
 s:ShortString; // 255
 c:char;
 i:integer;
 r:string; // reply
begin
 if busy then exit;
 busy:=true;

 i:=Telnet.GetData(ASocket,s[1],255);
 while i>0 do begin

  s[0]:=chr(i);
  r:='';
  for i:=1 to Length(s) do begin
   c:=s[i];
   case Status of

    stDATA: if c=tnIAC then Status:=stIAC else Data:=Data+c; // 5250 stream

    stIAC: case c of // IAC sequence
     tnIAC : begin Data:=Data+c; Status:=stDATA; end;
     tnDO  : Status:=stDO;
     tnSB  : begin Status:=stSB; SubNegoce:=''; end;
     tnSE  : ;
     tnWILL: Status:=stWill;
     tnEND : begin
              Status:=stDATA;
              ExFmt;
              Data:='';
             end;
     else begin Error('<!> IAC #'+IntToStr(ord(c))); exit; end;
    end;

    stDO: begin // DO - WILL/WONT
     Status:=stDATA;
     case c of
      tnBIN     : begin r:=r+tnIAC+tnWILL+tnBIN;       end;
      tnTERMTYPE: begin r:=r+tnIAC+tnWILL+tnTERMTYPE;  end;
      tnEOR     : begin r:=r+tnIAC+tnWILL+tnEOR;       end;
      tnNEWENV  : begin r:=r+tnIAC+tnWILL+c;           end;
      tnTIMING  : begin r:=r+tnIAC+tnWONT+tnTIMING;    end;
      else begin
       Error('<!> IAC DO +#'+IntToStr(Ord(c)));
       exit;
      end;
     end;
    end;

    stWILL: begin   // WILL - DO/DONT
     Status:=stDATA;
     case c of
      tnBIN        : begin r:=r+tnIAC+tnDO+tnBIN;         end;
      tnECHO       : begin r:=r+tnIAC+tnDO+tnECHO;        end;
      tnSUPPRESS_GA: begin r:=r+tnIAC+tnDO+tnSUPPRESS_GA; end;
      tnEOR        : begin r:=r+tnIAC+tnDO+tnEOR;         end;
      tnNEWENV     : begin r:=r+tnIAC+tnDO+tnNEWENV;      end;
      tnTIMING     : begin r:=r+tnIAC+tnDONT+tnTIMING;    end;
      else begin
       Error('<!> IAC WILL +#'+IntToStr(Ord(c)));
       exit;
      end;
     end;
    end;

   // Sub-negoce
    stSB: begin
     SubNegoce:=SubNegoce+c;
     if c=tnSE then begin
      if SubNegoce=tnTERMTYPE+ttSEND+tnIAC+tnSE then begin
       r:=r+tnIAC+tnSB+tnTERMTYPE+ttIS+TERM+tnIAC+tnSE;
      end else
      if SubNegoce[1]=tnNEWENV then begin
      { #39    #01  #03 #73#66#77#82#83#69#69#68 #127#13#212#104#179#11#04#04#00 #03 #255 }
      { NEWENV SEND #03 "IBMRSEED12345678"#0 #03               ?????????????????????????????????? IAC  }
       r:=r+tnIAC+tnSB+tnNEWENV+ttIS;
       r:=r+tnVAR+'DEVNAME'+tnVAL+DevName.Text;
       r:=r+tnVAR+'IBMMSGQNAME'+tnVAL+MSGQName.Text;
       r:=r+tnVAR+'IBMMSGQLIB' +tnVAL+MSGQLib.Text;
       r:=r+tnVAR+'IBMTRANSFORM'+tnVAL+'1';
       r:=r+tnVAR+'IBMMFRTYPMDL'+tnVAL+TypeModel.Text;
       if TypeModel.Text='*WSCST' then begin
        r:=r+tnVAR+'IBMWSCSTNAME'+tnVAL+WSCSTName.Text;
        r:=r+tnVAR+'IBMWSCSTLIB' +tnVAL+WSCSTLib.Text;
       end;
       r:=r+tnVAR+'IBMFONT'     +tnVAL+'12';
       r:=r+tnVAR+'IBMFORMFEED' +tnVAL+'C';
       r:=r+tnVAR+'IBMPPRSRC1'  +tnVAL+PaperSource(PPS1);//#4; // *A4
       r:=r+tnVAR+'IBMPPRSRC2'  +tnVAL+PaperSource(PPS2);//tnIAC+#255; // *NONE
       r:=r+tnVAR+'IBMENVELOPE' +tnVAL+EnvSource(ENV); // *NONE
       r:=r+tnIAC+tnSE;
      end else begin
       Error('<!> SubNegoce #'+IntToStr(ord(SubNegoce[1]))+ ' '+SubNegoce);
       exit;
      end;
      Status:=stDATA;
     end;
    end;

   // unsupported ?
    else begin
     Error('IAC #'+IntToStr(ord(c)));
     exit;
    end;

   end; //  case s[i]
  end; //  For s[i]

  if r<>'' then Telnet.Write(ASocket,r);
  i:=Telnet.GetData(ASocket,s[1],255);

 end;
 busy:=false;
end;


procedure TForm1.Error(Msg:string);
 begin
  Telnet.Active:=False;
  ShowMessage(Msg);
 end;

procedure TForm1.ExFmt;

 procedure StartupInfo;
  var
   x:integer;
   s:string;
  begin
   x:=ord(data[1])+2;
   s:=EbcdicToAnsiW(Copy(Data,x+4,4));
   Caption:=Trim(EbcdicToAnsiW(Copy(Data,x+16,8)))+'/'+Trim(EbcdicToAnsiW(Copy(Data,x+8,8)));
   if s[1]='I' then s[1]:='1';
   case StrToInt(s) of
    1901 : PrinterStatusCaption:='Virtual device has less function than source device.';
    1902 : PrinterStatusCaption:='Session successfully started';
    1906 : PrinterStatusCaption:='Automatic sign-on requested, but not allowed.';
    2702 : PrinterStatusCaption:='Device description not found.';
    2703 : PrinterStatusCaption:='Controller description not found.';
    2777 : PrinterStatusCaption:='Damaged device description.';
    8901 : PrinterStatusCaption:='Device not varied on.';
    8902 : PrinterStatusCaption:='Device not available.';
    8903 : PrinterStatusCaption:='Device not valid for session';
    8906 : PrinterStatusCaption:='Session initiation failed.';
    8907 : PrinterStatusCaption:='Session failure.';
    8910 : PrinterStatusCaption:='Controller not valid for session.';
    8916 : PrinterStatusCaption:='No matching device found.';
    8917 : PrinterStatusCaption:='Not authorized to object.';
    8918 : PrinterStatusCaption:='Job canceled.';
    8920 : PrinterStatusCaption:='Object partially damaged.';
    8921 : PrinterStatusCaption:='Communications error.';
    8922 : PrinterStatusCaption:='Negative response received.';
    8923 : PrinterStatusCaption:='Start-up record built incorrectly.';
    8925 : PrinterStatusCaption:='Creation of device failed.';
    8928 : PrinterStatusCaption:='Change of device failed.';
    8929 : PrinterStatusCaption:='Vary on or vary off failed.';
    8930 : PrinterStatusCaption:='Message queue does not exist';
    8934 : PrinterStatusCaption:='Start-up for S/36 WSF received.';
    8935 : PrinterStatusCaption:='Session rejected.';
    8936 : PrinterStatusCaption:='Security failure on session attempt.';
    8937 : PrinterStatusCaption:='Automatic sign-on rejected.';
    8940 : PrinterStatusCaption:='Automatic configuration failed or not allowed.';
    1904 : PrinterStatusCaption:='Source system at incompatible release.';
    else   PrinterStatusCaption:='Unknow error #'+s;
   end;
   {$IFDEF STREAM}
   Stream:=nil;
   number:=0;
   {$ENDIF}
   {$IFDEF PRINT}
   hPrinter:=0;
   {$ENDIF}
   ibmcode:=0;
  end;

 procedure PrintComplet;
  var
   s:string;
  begin
   s:=#$00#$0A#$12#$A0#$01#$02#$04#$00#$00#$01#$FF#$EF;
{
   Figure 5 shows the client sending the server a print complete record.
   This would normally follow receipt of a print record, such as the
   example in Figure 4.  This indicates successful completion of a print
   request.

   +--------------------------------------------------------------------+
   |   +-- Length of structure (LLLL)                                   |
   |   |    +-- GDS identifier                                          |
   |   |    |    +-- Data flow record                                   |
   |   |    |    |   +-- Length of pass-through specific header (LL)    |
   |   |    |    |   |   +-- Flags                                      |
   |   |    |    |   |   |   +-- Printer operation code                 |
   |   |    |    |   |   |   |                                          |
   | +--+ +--+ +--+ ++ +--+ ++                                          |
   | |  | |  | |  | || |  | ||                                          |
   | 000A 12A0 0102 04 0000 01                                          |
   +--------------------------------------------------------------------+
    Figure 5. Client sending server a print complete record

   - '000A'X = Logical record length, including this byte (LLLL)
   - '12A0'X = GDS LU6.2 header
   - '0102'X = Data flow response record (client to server)
   - '04'X   = Length of pass-through specific header (LL)
   - '0000'X = Good Response
   - '01'X   = Print Complete
 }
   Telnet.SendData(s[1],Length(s));
  end;

 procedure PrintError;
  var
   s:string;
  begin
   DataFlow.Brush.Color:=clRed;
   PrinterStatusCaption:='Local printer error';
   s:=#$00#$0A#$12#$A0#$01#$02#$04#$81#$81#$01#$FF#$EF; // Device Error ???!!!
   Telnet.SendData(s[1],Length(s));
  end;

 Procedure PrintRecord;
 var
  x:integer;
  i:integer;
  error:boolean;
 begin
  error:=false;
  x:=ord(Data[1])+1; // data offset
  while x<=length(data) do begin // until end of buffer
   if (ibmcode=0)and(data[x]=#0) then begin // found #0 outside a escape sequence
    inc(x);
    {$IFDEF STREAM}
    if Stream=nil then
     ShowMessage('<!> Stream = nil !!!')
    else begin
     Stream.SaveToFile('print'+IntToStr(number)+'.pcl');
     Stream.Free;
     Stream:=nil;
    end;
    {$ENDIF}
    {$IFDEF PRINT}
     EndDoc(hPrinter);
    {$ENDIF}
    PrinterStatusCaption:='READY.';
    DataFlow.Brush.Color:=clBlack;
   end else begin
    if ibmcode=0 then begin
     if data[x]<>#3 then ShowMessage('0x03 not found :(');  inc(x);
     ibmcode:=ord(data[x]); inc(x);
     hpcl:='';
    end;
    while (x<=length(data))and(ibmcode>0) do begin
     hpcl:=hpcl+data[x]; inc(x);
     dec(ibmcode);
     if ibmcode=0 then begin
     {$IFDEF STREAM}
      if Stream=nil then begin
       Stream:=TMemoryStream.Create;
       inc(number);
      end;
      Stream.WriteBuffer(pchar(hpcl),Length(hpcl));
     {$ENDIF}
     {$IFDEF PRINT}
      if hPrinter=0 then begin
       hPrinter:=OpenPrinter(Printers.Text);
       PrinterStatusCaption:='Printing...';
       DataFlow.Brush.Color:=clGreen;
       DataFlow.Update;
       BeginDoc(hPrinter,DevName.Text+'@'+HostName.Text);
      end;
      if not WritePrinter(hPrinter,hpcl) then begin
       Error:=True;
       AbortDoc(hPrinter);
      end;
     {$ENDIF}
     end;
    end;
   end;
  end;
  if error then PrintError else PrintComplet;
 end;

 begin
  Data2:=Data;
  if Length(Data)<SizeOf(Head) then begin
   showMessage('Invalide data !');
   exit;
  end;
  Move(Data[1],Head,SizeOf(Head));
  Delete(Data,1,SizeOf(Head));

  Case Head.Flow of
    $0090 : StartupInfo;
    $0101 : PrintRecord;
   else begin
    Error('Unknow data flow #'+IntToStr(Head.flow));
    exit;
   end;
  end;
 end;

procedure TForm1.TelnetClose(Sender: TObject; ASocket: Integer);
begin
 StopClick(Self);
 if PrinterStatus.Caption='Session successfully started' then
  PrinterStatusCaption:='Diconnected by host';
end;

procedure TForm1.TelnetError(Sender: TObject; Error: Word;
  ASocket: Integer; Event: Word);
begin
 PrinterStatusCaption:=('SOCKET ERROR #'+IntToStr(Error));
 //Telnet.Active:=false;
 StopClick(Self);
end;

procedure TForm1.ListPrinters;
 const
  Flags=PRINTER_ENUM_LOCAL;
  Level=5;
 var
  Count, NumInfo: Integer;
  Buffer, PrinterInfo: PChar;
  i:integer;
 begin
  EnumPrinters(Flags, nil, Level, nil, 0, Count, NumInfo);
  GetMem(Buffer, Count);
  try
   if not EnumPrinters(Flags, nil, Level, PByte(Buffer), Count, Count, NumInfo) then Exit;
   PrinterInfo := Buffer;
   for I := 0 to NumInfo - 1 do begin
    with PPrinterInfo5(PrinterInfo)^ do begin
     Printers.Items.Add(pPrinterName);
     Inc(PrinterInfo, sizeof(TPrinterInfo5));
    end
   end;
  finally
   FreeMem(Buffer, Count);
  end;
 end;

function DefaultPrinter:string;
 var
  pDefaultPrinter: array[0..79] of Char;
  i:integer;
 begin
  GetProfileString('windows', 'device', '', pDefaultPrinter,SizeOf(pDefaultPrinter) - 1);
  Result:=pDefaultPrinter;
  i:=pos(',',Result);
  if i>0 then SetLength(Result,i-1);
 end;

procedure SetItem(List:TComboBox;Index:integer);
 begin
  Index:=List.Items.IndexOf(ParamStr(Index));
  if Index>=0 then List.ItemIndex:=Index;
 end;

procedure TForm1.FormCreate(Sender: TObject);
var
 x:integer;
 s:string;
begin
// Windows Printers
 ListPrinters;
 Printers.ItemIndex:=Printers.Items.IndexOf(DefaultPrinter);
// Default
 TypeModel.ItemIndex:=0;
 PPS1.ItemIndex:=0;
 PPS2.Items:=PPS1.Items;
 PPS2.ItemIndex:=0;
 ENV.ItemIndex:=0;
// Parms
 if ParamCount=0 then
  MoreClick(Self)
 else begin
  x:=0;
  SetItem(Printers,x+1);
  HostName.Text:=ParamStr(x+2);
  DevName.Text:=ParamStr(x+3);
  TypeModel.Text:=ParamStr(x+4);
  inc(x,4);
  if TypeModel.Text='*WSCST' then begin
   WSCSTName.Text:=ParamStr(x+1);
   WSCSTLib.Text :=ParamStr(x+2);
   inc(x,2);
  end;
  SetItem(PPS1,x+1);
  SetItem(PPS2,x+2);
  SetItem(ENV,x+3);
  LessClick(Self);
  StartClick(Self);
 end;
end;

procedure TForm1.wmToolTrayIcon(Var msg:TMessage);
 begin
  case msg.lParam of
   WM_LBUTTONDBLCLK : if Not Visible then Show else Hide;
//   WM_RBUTTONUP     : Popup;
  end;
 end;

procedure TForm1.StartClick(Sender: TObject);
begin
 if HostName.Text='' then begin
  MoreClick(Self);
  HostName.SetFocus;
 end else begin
  PrinterStatusCaption:='Connexion...';
  DataFlow.Brush.color:=clBlack;
  Telnet.Call(HostName.Text);
  Start.Enabled:=False;
 end;
end;

procedure TForm1.StopClick(Sender: TObject);
begin
 OnLine.Brush.color:=clBlack;
 Telnet.Active:=False;
 Start.Caption:='Start';
 Start.OnClick:=StartClick;
 Start.Enabled:=True;
 Caption:='Telnet/400';
 if Sender=Start then PrinterStatusCaption:='Offline';
end;

procedure TForm1.AboutClick(Sender: TObject);
begin
 ShowMessage('TN3812 v3.0, (C)1999 by Paul TOTH'#13#10+
             'http://www.multimania.com/tothpaul/home.shtml'#13#10+
             #13#10+
             'This program is free software; you can redistribute it and/or'#13#10+
             'modify it under the terms of the GNU General Public License'#13#10+
             'as published by the Free Software Foundation; either version 2'#13#10+
             'of the License, or (at your option) any later version.'#13#10+
             #13#10+
             'This program is distributed in the hope that it will be useful,'#13#10+
             'but WITHOUT ANY WARRANTY; without even the implied warranty of'#13#10+
             'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the'#13#10+
             'GNU General Public License for more details.'#13#10+
             #13#10+
             'You should have received a copy of the GNU General Public License'#13#10+
             'along with this program; if not, write to the Free Software'#13#10+
             'Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.');

end;

procedure TForm1.MoreClick(Sender: TObject);
begin
 Panel.Enabled:=True;
 ClientHeight:=Panel.Left+Panel.Top+Panel.Height;
 More.Caption:='<< Less';
 More.OnClick:=LessClick;
end;

procedure TForm1.LessClick(Sender: TObject);
begin
 ClientHeight:=Panel.Top;
 More.Caption:='More >>';
 More.OnClick:=MoreClick;
 Panel.Enabled:=False;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
 Telnet.Active:=False;
end;

procedure TForm1.SetStatus(s:string);
 begin
  PrinterStatus.Hint:=s;
  PrinterStatus.Caption:=s;
 end;

end.

