unit sockets;

{
  Socket component for Delphi (c)1999 by Paul TOTH <tothpaul@multimania.com>
  http://www.multimania.com/tothpaul

  fix : TCustomSocket renamed to TPaulSocket to avoid Delphi 4 conflict.

}

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

{ WINSOCK support }
Const
 WinSock='wsock32.dll'; { 32bits socket DLL }

Type
  TWSAData = packed record
    wVersion: Word;
    wHighVersion: Word;
    szDescription: array[0..256] of Char;
    szSystemStatus: array[0..128] of Char;
    iMaxSockets: Word;
    iMaxUdpDg: Word;
    lpVendorInfo: PChar;
  end;

  TSockAddr=packed record
   Family:word;
   Port:word;
   Addr:LongInt;
   Zeros:array[0..7] of byte;
  end;

type
  TFamily=(fmInternet);
  TPortName=(poNull,poFTP,poTELNET,poSMTP,poHTTP,poPOP3,poNNTP,poOther);
  TEvents=set of (evRead,evWrite,evOOB,evAccept,evConnect,evClose);

  TSocketEvent=procedure(Sender:TObject; ASocket:Integer) of object;
  TErrorEvent=procedure(Sender:TObject; Error:word; ASocket:Integer; Event:word) of object;

Const
 FamilyValue:array[TFamily] of Integer = (2);
 scStream=1;
 SetNonBlocking = $8004667E; { $80 , SizeOf(LongInt) , 'f' , 126 }

Type
  TPaulSocket = class(TComponent)
  private
   fFamily:TFamily;
   fPort:word;
   fIpAddr:string;
   fActive:boolean;
   fBlockingMode:boolean;
   fEvents:TEvents;
   fServer:boolean;
   EOnRead:TSocketEvent;
   EOnWrite:TSocketEvent;
   EOnConnect:TSocketEvent;
   EOnAccept:TSocketEvent;
   EOnOOB:TSocketEvent;
   EOnClose:TSocketEvent;
   EOnError:TErrorEvent;
   AutoStart:boolean;
   procedure SetActive(Value:boolean);
   procedure SetEvents(Value:TEvents);
   procedure SetServer(Value:boolean);
   procedure WndProc(Var Msg:TMessage);
  protected
   Started:boolean;
   HSocket:Integer; { should be a word, with $FFFF as invalid socket }
   Handle:THandle;
   fBuffer:Array[0..1024] of char;
   fTextBuf:Array[0..1024] of char;
   fTextPos:word;
   procedure DoRead(ASocket:integer);
   procedure DoWrite(ASocket:integer);
   procedure DoOOB(ASocket:integer);
   procedure DoAccept(ASocket:integer);
   procedure DoConnect(ASocket:integer);
   procedure DoClose(ASocket:integer);
   procedure DoError(ASocket:integer; Error:word; Event:word);
   procedure StartServer;
  public
   Function SockInfo:TSockAddr;
   Function AcceptSocket:integer;

   constructor Create(AOwner:TComponent); override;
   destructor Destroy; override;
   procedure Loaded; override;
   procedure Call(AHost:string);
   function GetData(ASocket:integer; Var Data; DataLen:word):integer;
   function PutData(ASocket:integer; Var Data; DataLen:word):integer;
   function Write(ASocket:integer; s:string):integer;
   function WriteLn(ASocket:integer; s:string):integer;
   function SendData(Var Data; DataLen:integer):integer;
   function Echo(s:string):integer;
   function ReadLn(ASocket:integer; var s:string):integer;
   function Read(var s:string):integer;
   procedure Close(ASocket:integer);
//  published
   Property Family:TFamily read fFamily write fFamily default fmInternet;
   Property Port:word read fPort write fPort;
   Property IpAddr:string read fIpAddr write fIpAddr;

   Property Active:boolean read fActive write SetActive;
   Property Events:TEvents read fEvents write SetEvents;
   Property Server:boolean read fServer write SetServer;

   Property BlockingMode:boolean read fBlockingMode write fBlockingMode;

   Property OnRead:TSocketEvent read EOnRead write EOnRead;
   Property OnWrite:TSocketEvent read EOnWrite write EOnWrite;
   Property OnAccept:TSocketEvent read EOnAccept write EOnAccept;
   Property OnConnect:TSocketEvent read EOnConnect write EOnConnect;
   Property OnClose:TSocketEvent read EOnClose write EOnClose;
   Property OnOOB:TSocketEvent read EOnOOB write EOnOOB;

   Property OnError:TErrorEvent read EOnError write EOnError;
  end;

  TSocket=Class(TPaulSocket)
  private
   procedure SetPortName(Value:TPortName);
   function  GetPortName:TPortName;
  published
   Property Active;
   Property Events;
   Property Server;

   Property Family;
   Property IpAddr;
   Property Port;
   Property PortName:TPortName read GetPortName write SetPortName;

   Property BlockingMode;

   Property OnRead;
   Property OnWrite;
   Property OnAccept;
   Property OnConnect;
   Property OnClose;
   Property OnOOB;
   Property OnError;
  end;

procedure Register;
Function LongToIp(Long:LongInt):string;
Function LocalName:string;
Function LocalIp:string;
Function HostToLong(AHost:string):LongInt;

implementation

{ Register }

procedure Register;
begin
  RegisterComponents('MySoft', [TSocket]);
end;

{ Winsock }
Function WSAStartup(Version:word; Var Data:TwsaData):integer; stdcall; far; external winsock;
Function socket(Family,Kind,Protocol:integer):integer; stdcall; far; external winsock;
Function shutdown(Socket,How:Integer):integer; stdcall; far; external winsock;
Function closesocket(socket:Integer):integer; stdcall; far; external winsock;
Function WSACleanup:integer; stdcall; far; external winsock;
Function WSAAsyncSelect(Socket:Integer; Handle:Hwnd; Msg:word; Level:Longint):longint; stdcall; far; external winsock;
Function bind(Socket:Integer; Var SockAddr:TSockAddr; AddrLen:integer):integer; stdcall; far; external winsock;
Function listen(socket,flags:Integer):integer; stdcall; far; external winsock;
Function connect(socket:Integer; Var SockAddr:TSockAddr; AddrLen:integer):integer; stdcall; far; external winsock;
Function accept(socket:Integer; Var SockAddr:TSockAddr; Var AddrLen:Integer):integer; stdcall; far; external winsock;
Function WSAGetLastError:integer; stdcall; far; external winsock;
Function recv(socket:integer; var data; datalen,flags:integer):integer; stdcall; far; external winsock;
Function send(socket:integer; var data; datalen,flags:integer):integer; stdcall; far; external winsock;
//Function getpeername(socket:integer; var SockAddr:TSockAddr; Var AddrLen:Integer):Integer; stdcall; far; external winsock;
Function gethostname(var name; maxlen:integer):integer; stdcall; far; external winsock;
Function gethostbyname(HostName:PChar):pointer; stdcall; far; external winsock;
Function getsockname(socket:integer; var SockAddr:TSockAddr; Var AddrLen:Integer):integer; stdcall; far; external winsock;

Function inet_ntoa(addr:longint):PChar; stdcall; far; external winsock;

Function WSAIsBlocking:boolean; stdcall; far; external winsock;
Function WSACancelBlockingCall:integer; stdcall; far; external winsock;

Function ioctlsocket(socket:integer; cmd: Longint; var arg: longint): Integer; stdcall; far; external winsock;

//
Var
 Started:boolean;
 wsaData:TWSAData;

Function Startup:boolean;
 begin
  Started:=(Started)or(wsaStartup($101,wsaData)=0);
  Result:=Started;
 end;

Procedure Cleanup; far;
 begin
  if Started then begin
   wsaCleanup;
   Started:=False;
  end;
 end;

Function IpToLong(ip:string):LongInt;
 var
  x,i:byte;
  ipx:array[0..3] of byte;
  v:integer;
 begin
  Result:=0;
  longint(ipx):=0; i:=0;
  for x:=1 to length(ip) do
   if ip[x]='.' then begin
    inc(i);
    if i=4 then exit;
   end else begin
    if not (ip[x] in ['0'..'9']) then exit;
    v:=ipx[i]*10+ord(ip[x])-ord('0');
    if v>255 then exit;
    ipx[i]:=v;
   end;
  result:=longint(ipx);
 end;

Function HostToLong(AHost:string):LongInt;
 Type
  PHostEnt = ^THostEnt;
  THostEnt = packed record
    name     : PChar;
    aliases  : ^PChar;
    addrtype : Smallint;
    length   : Smallint;
    addr     : ^pointer;
  end;
 Var
  Host:PHostEnt;
 begin
  Host:=GetHostByName(PChar(AHost));
  if Host=nil then
   Result:=IpToLong(AHost)
  else
   Result:=LongInt(Host^.Addr^^);
 end;

Function LongToIp(Long:LongInt):string;
 var
  ipx:array[0..3] of byte;
  i:byte;
 begin
  longint(ipx):=long;
  Result:='';
  for i:=0 to 3 do result:=result+IntToStr(ipx[i])+'.';
  SetLength(Result,Length(Result)-1);
 end;

Function LocalName:string;
 Var
  Name:array[0..50] of char;
 begin
  if gethostname(Name,SizeOf(Name))=0 then
   Result:=Name
  else
   Result:='';
 end;

Function LocalIp:string;
 begin
  if Startup then
   Result:=LongToIp(HostToLong(LocalName))
  else
   Result:='';
 end;

{ TSocket }
Constructor TPaulSocket.Create(AOwner:TComponent);
 begin
  inherited Create(AOwner);
  Handle:=AllocateHWnd(WndProc);
 end;

Destructor TPaulSocket.Destroy;
 begin
  Active:=false;
  DeallocateHWnd(Handle);
  inherited Destroy;
 end;

Procedure TPaulSocket.Loaded;
 begin
  inherited Loaded;
  if AutoStart then Active:=True;
 end;

Procedure TPaulSocket.WndProc(Var Msg:TMessage);
 begin
  with msg do
   if msg=wm_user then
    if LParamHi=0 then begin
     case lParamLo of
      $0001 : DoRead(wParam);
      $0002 : DoWrite(wParam);
      $0004 : DoOOB(wParam);
      $0008 : DoAccept(wParam);
      $0010 : DoConnect(wParam);
      $0020 : DoClose(wParam);
     end;
    end else
      DoError(wParam,LParamHi,LParamLo)
   else
    Result:=DefWindowProc(Handle, Msg, wParam, lParam);
 end;

Procedure TPaulSocket.DoRead(ASocket:integer);
 begin
  if Assigned(EOnRead) then EOnRead(Self,ASocket);
 end;

Procedure TPaulSocket.DoWrite(ASocket:integer);
 begin
  if Assigned(EOnWrite) then EOnWrite(Self,ASocket);
 end;

Procedure TPaulSocket.DoOOB(ASocket:integer);
 begin
  if Assigned(EOnOOB) then EOnOOB(Self,ASocket);
 end;

Procedure TPaulSocket.DoAccept(ASocket:integer);
 var
  SockAddr:TSockAddr;
  SockAddrLen:integer;
 begin
  SockAddrLen:=SizeOf(TSockAddr);
  ASocket:=Accept(ASocket,SockAddr,SockAddrLen);
  if Assigned(EOnAccept) then EOnAccept(Self,ASocket);
 end;

Procedure TPaulSocket.DoConnect(ASocket:integer);
 begin
  { Default ip Address (for FTP client) }
  fIpAddr:=LongToIp(SockInfo.Addr);

  if Assigned(EOnConnect) then EOnConnect(Self,ASocket);
 end;

Procedure TPaulSocket.DoClose(ASocket:integer);
 begin
  if Assigned(EOnClose) then EOnClose(Self,ASocket);
  ShutDown(ASocket,2);
  CloseSocket(Asocket);
 end;

Procedure TPaulSocket.DoError(ASocket:integer; Error:word; Event:word);
 begin
  if Error=10035 then exit; { WSAEWOULDBLOCK }
  if Assigned(EOnError) then EOnError(Self,Error,ASocket,Event);
 end;

Procedure TPaulSocket.SetActive(Value:boolean);
 Const
  SetOn:longint=1;
 begin
  if fActive=value then exit;
  if Value then begin
  { While loading ? }
   if csLoading in ComponentState then begin
    AutoStart:=True;
    exit;
   end;
  { init Winsock DLL }
   Started:=(Started) or (Startup);
   if not Started then exit;
  { Get socket }
   HSocket:=Socket(FamilyValue[fFamily],scStream,0);
   if HSocket=-1 then exit;
  { allright }
   fActive:=True;

  { Listen on this socket }
   if not fBlockingMode then Events:=fEvents else ioctlsocket(HSocket,SetNonBlocking,SetOn);

  { Start Server }
   StartServer;
  end else begin
  { BlockMode }
   if wsaIsBlocking then wsaCancelBlockingCall;
  { free socket }
   if HSocket<>-1 then begin
    ShutDown(HSocket,2);
    CloseSocket(HSocket);
    HSocket:=-1;
   end;
  { done Winsock DLL }
   Started:=False;
   Cleanup;
  { done }
   fActive:=False;
  end;
 end;

Procedure TPaulSocket.SetEvents(Value:TEvents);
 begin
  fEvents:=Value;
  if (not Active)or(HSocket=-1) then exit;
  wsaAsyncSelect(HSocket,Handle,wm_user,LongInt(Byte(fEvents)));
 end;

Procedure TPaulSocket.SetServer(Value:boolean);
 begin
  fServer:=Value;
  if (csLoading in ComponentState) then exit;
  if Value then StartServer else begin
  { recreate HSocket }
   Active:=False;
   Active:=True;
  end;
 end;

Procedure TPaulSocket.StartServer;
 var
  SockAddr:TSockAddr;
  name:array[0..20] of char;
 begin
  if Active and Server then begin
   FillChar(SockAddr,SizeOf(SockAddr),0);
   SockAddr.Family:=FamilyValue[fFamily];
   SockAddr.Port:=swap(fPort);
   // retreive ip address
   if (fIpAddr='')and(gethostname(name,sizeof(name))=0) then fIpAddr:=LongToIp(HostToLong(name));
   SockAddr.Addr:=IpToLong(fIpAddr);
   if Bind(HSocket,SockAddr,SizeOf(SockAddr))<>0 then
    DoError(HSocket,wsaGetLastError,0)
   else
    if Listen(HSocket,0)<>0 then DoError(HSocket,wsaGetLastError,0);
  end;
 end;

Procedure TPaulSocket.Call(AHost:String);
 Var
  SockAddr:TSockAddr;
 begin
  Active:=true;
  fTextBuf[0]:=#0;
  FillChar(SockAddr,SizeOf(SockAddr),0);
  SockAddr.Family:=FamilyValue[fFamily];
  SockAddr.Port:=swap(fPort);
  SockAddr.Addr:=HostToLong(AHost);
{
  if SockAddr.Addr=0 then begin
   SockAddr.Addr:=HostToLong(AHost);
  end;
}  
  if Connect(HSocket,SockAddr,SizeOf(SockAddr))<>0 then DoError(HSocket,wsaGetLastError,0);
 end;

Function TPaulSocket.SockInfo:TSockAddr;
 var
  SockLen:integer;
 begin
  SockLen:=SizeOf(Result);
  FillChar(Result,Sizeof(Result),0);
  GetSockName(HSocket,Result,SockLen);
 end;
{
Function TPaulSocket.GetIpAddr:string;
 Var
  SockAddr:TSockAddr;
  SockLen:Integer;
 begin
  SockLen:=SizeOf(SockAddr);
  if GetSockName(HSocket,SockAddr,SockLen)=0 then
   Result:=LongToIp(SockAddr.Addr)
  else
   Result:='???.???.???.???';
 end;

Function TPaulSocket.GetPort:word;
 Var
  SockAddr:TSockAddr;
  SockLen:Integer;
 begin
  SockLen:=SizeOf(SockAddr);
  if GetSockName(HSocket,SockAddr,SockLen)=0 then
   Result:=SockAddr.Port
  else
   Result:=0;
 end;
}
Function TPaulSocket.GetData(ASocket:integer;Var Data; DataLen:word):integer;
 begin
  Try
   Result:=recv(ASocket,Data,DataLen,0);
   if Result<0 then begin
    DoError(ASocket,wsaGetLastError,word(evRead));
    Result:=0;
   end;
  Except
   Result:=0; //- nothing on error ...
  End;
 end;

Function TPaulSocket.PutData(ASocket:integer; Var Data; DataLen:word):integer;
 begin
//  Move(Data,fBuffer,DataLen);
//  Result:=send(ASocket,fBuffer,DataLen,0);
  Result:=Send(ASocket,Data,DataLen,0);
  if Result=-1 then begin
   DoError(ASocket,wsaGetLastError,word(evWrite));
   Result:=0;
  end;
 end;

Function TPaulSocket.Write(ASocket:integer; s:string):integer;
 begin
  Result:=PutData(ASocket,PChar(s)^,Length(s));
 end;

Function TPaulSocket.WriteLn(ASocket:integer; s:string):integer;
 begin
  Result:=Write(ASocket,s+#13#10);
 end;

Function TPaulSocket.SendData(Var Data; DataLen:integer):integer;
 begin
  Result:=PutData(HSocket,Data,DataLen);
 end;

Function TPaulSocket.Echo(s:string):integer;
 begin
  Result:=WriteLn(HSocket,s);
 end;

Function TPaulSocket.ReadLn(ASocket:integer; var s:string):integer;
 var
  CR:PChar;
  more:integer;

 Function GetCR:Pchar;
  Var
   LF:PChar;
  begin
   Result:=StrPos(fTextBuf,#13);
   if Result=nil then
    Result:=StrPos(fTextBuf,#10)
   else begin
    Result^:=#0;
    LF:=StrPos(fTextBuf,#10);
    if LF<>nil then begin
     Result^:=#13;
     Result :=LF;
   end;
  end;
 end;
 
 Function FillBuffer:integer;
  begin
   fTextPos:=StrLen(fTextBuf);
   Result:=GetData(ASocket,fTextBuf[fTextPos],SizeOf(fTextBuf)-fTextPos-1);
   inc(fTextPos,Result);
   fTextBuf[fTextPos]:=#0; // as PChar
   CR:=GetCR;
  end;

 begin
  s:=''; Result:=-1;
  CR:=GetCR;
  if (CR=nil)and(FillBuffer=0) then exit;
  While CR=nil do begin // this is a more_then _1024_chars line !!!
   if fTextPos<1024 then exit; // warning ! the line can be incomplet !!!
   s:=fTextBuf;
   fTextbuf[0]:=#0;
   if FillBuffer=0 then begin // nothing more to read
    Result:=Length(s); // return what we read
    exit;
   end;
   CR:=GetCR;
  end;
  CR^:=#0;
  s:=s+fTextBuf;
  inc(CR); if CR^ in [#10,#13] then inc(CR);
  StrCopy(fTextBuf,CR);
  Result:=Length(s);
 end;

Function TPaulSocket.read(var s:string):integer;
 begin
  result:=ReadLn(HSocket,s);
 end;

Procedure TPaulSocket.close(ASocket:integer);
 begin
  DoClose(ASocket);
 end;

Function TPaulSocket.AcceptSocket:integer;
 var
  SockAddr:TSockAddr;
  SockAddrLen:integer;
 begin
  FillChar(SockAddr,SizeOf(TSockAddr),0);
  SockAddrLen:=SizeOf(TSockAddr);
  Result:=Accept(HSocket,SockAddr,SockAddrLen);
 end;

// SOCKET
const
 PortValue:array[TPortName] of word = (0,21,23,25,80,110,119,0);

Procedure TSocket.SetPortName(Value:TPortName);
 begin
  if (not Active)and(Value<poOther) then fPort:=PortValue[Value];
 end;

Function TSocket.GetPortName:TPortName;
 begin
  case fPort of
   0  : Result:=poNull;
   21 : Result:=poFTP;
   23 : Result:=poTELNET;
   25 : Result:=poSMTP;
   80 : Result:=poHTTP;
   110: Result:=poPOP3;
   119: Result:=poNNTP;
   else Result:=poOther;
  end;
 end;

begin
 Started:=false;
 AddExitProc(Cleanup);
end.
